#ifndef _OLWM_LBL_CURSORS_H
#define _OLWM_LBL_CURSORS_H

#ifdef IDENT
#ident "@(#)cursors.h	1.2 olvwm version 25 Nov 1992"
#endif

extern void InitCursors( /* Display *,  ScreenInfo * */ );
extern void updateCursors( /* Display *, char * */ );

#endif
