#ifdef IDENT
#ident	"@(#)defaults.h	26.16	93/06/28 SMI"
#endif

/*
 *      (c) Copyright 1989 Sun Microsystems, Inc.
 */

/*
 *      Sun design patents pending in the U.S. and foreign countries. See
 *      LEGAL_NOTICE file for terms of the license.
 */

#ifndef _OLWM_DEFAULTS_H
#define _OLWM_DEFAULTS_H

extern void GetDefaults();
extern XrmDatabase GetUserDefaults();
extern XrmDatabase GetAppDefaults();

#endif /* _OLWM_DEFAULTS_H */
