#ifdef IDENT
#ident	"@(#)olcursor.h	26.7	93/06/28 SMI"
#endif

/*
 *      (c) Copyright 1989 Sun Microsystems, Inc.
 */

/*
 *      Sun design patents pending in the U.S. and foreign countries. See
 *      LEGAL_NOTICE file for terms of the license.
 */

#ifndef _OLWM_OLCURSOR_H
#define _OLWM_OLCURSOR_H

#define OLC_basic 0
#define OLC_move 2
#define OLC_copy 4
#define OLC_busy 6
#define OLC_stop 8
#define OLC_panning 10
#define OLC_target 12
#define OLC_nouse 14
#define OLC_ptr 16
#define OLC_beye 18
#define OLC_rtarr 20
#define OLC_xhair 22
#define OLC_xcurs 24
#define OLC_hourg 26

#endif /* _OLWM_OLCURSOR_H */
